import base64
import csv
import re                            
from cStringIO import StringIO
from AccessControl import ClassSecurityInfo
from Globals import InitializeClass
from OFS.SimpleItem import SimpleItem
from Products.CMFCore import CMFCorePermissions
from Products.CMFCore.utils import UniqueObject
from Products.CMFCore.utils import getToolByName

from config import CSV_KEY

class MemberImportTool(UniqueObject, SimpleItem):
    """ Member Import Tool.
    """

    id = 'member_import_tool'
    meta_type = 'Member IMport Tool'

    security = ClassSecurityInfo()
    security.declareObjectProtected(CMFCorePermissions.View)


    def _del_session_cache(self):
        ses = self.REQUEST.SESSION
        if ses.has_key(CSV_KEY):
            del ses[CSV_KEY]


    security.declareProtected(CMFCorePermissions.View,
        'uploadCSVFile')
    def uploadCSVFile(self, REQUEST=None):
        """
        Read uploaded csv file and save it in sessions variable
        """
        session = self.REQUEST.SESSION
        csv_file = self.REQUEST.get('csv_file', None)
        if csv_file is None:
            self._del_session_cache()
        else:
            session[CSV_KEY] = csv_file.read()

        if REQUEST:
            url = self.REQUEST.get('current_url', self.absolute_url())
            self.REQUEST.RESPONSE.redirect(
                '%s/mi_import_members_format' % url)


    def _site_encoding(self):
        "Returns the site encoding"
        portal_properties = self.portal_properties
        site_props = portal_properties.site_properties
        return site_props.default_charset or 'utf-8'


    security.declareProtected(CMFCorePermissions.View,
                              'csv_parser')
    def csv_parser(self, delimiter='comma', quote_char='double_quote',
                   coding='latin-1'):
        """Don't import HectoMegaByte files, or you are asking for it!"""
        delim_map = {'tab':'\t', 'semicolon':';', 'colon':':', 'comma':',', 'space':' ', }
        delimiter = delim_map[delimiter]
        quote_map = {'double_quote':'"', 'single_quote':"'", }
        quote_char = quote_map[quote_char]
        charset = self._site_encoding()
        csv_file_content = self.REQUEST.SESSION.get(CSV_KEY, '')

        # Horrid workaround to get the buffer in U (universal line terminator) mode
        from tempfile import mkstemp
        from os import close, write, system
        # Write temp file
        fp, fname = mkstemp()
        try:
            write(fp, csv_file_content)
        finally:
            close(fp)
        # Read temp file
        fh = open(fname, 'rU')
        try:
            buf = fh.read()
        finally:
            fh.close()
            system('rm "%s"' % fname)

        # this sucks! double memory usage
        f = StringIO(buf.decode(coding, 'replace').encode(charset))
        reader = csv.reader(f, delimiter=delimiter, lineterminator=chr(13))
        rows = []
        for row in reader:
            rows.append(row)
        return rows


    security.declareProtected(CMFCorePermissions.View,
                              'memberTypeFieldNames')
    def memberTypeFieldNames(self):
        "Returns the field names for member type"
        pmd = getToolByName(self, 'portal_memberdata')
        return ['username','password','roles'] + pmd.propertyIds()

    def _getFieldsMap(self, REQUEST):
        "Returns the fields map from the request"
        fields_map = {}
        n_fields = REQUEST.get('n_fields', 0)
        for n in range(n_fields):
            fieldname = REQUEST.get('fields_map_%s' % n, '')
            if fieldname != '*discard*':
                fields_map.setdefault(fieldname, []).append(n)
        return fields_map


    def _field_types(self):
        "Returns a dict with mapping of field and its type"
        pmd = getToolByName(self, 'portal_memberdata')
        ret = {'username':'string', 'password':'string', 'roles':'lines', 'domains':'lines'}
        for di in pmd.propertyMap():
            ret[di['id']] = di['type']

        return ret

    def _unique(self, dicts):
        "Removes duplicates from the import map."
        uniques = {}
        for dict_ in dicts:
            keys = dict_.keys()
            keys.sort()
            temp = []
            for key in keys:
                # stringify value in case we have an unhashable type
                val = str(dict_[key])
                temp.append((key, val))
            uniques[tuple(temp)] = dict_
        vals = uniques.values()
        return vals


    def _addMembers(self, default_role_is_member, import_maps, REQUEST=None):
        "Add members to site."
        pm = getToolByName(self, 'portal_membership')
        import_maps = self._unique(import_maps)
        now = self.ZopeTime()
        d_time = now.strftime('%Y%m%dZ%H%M%S')
        
        number_to_add = len(xrange(len(import_maps)))
        if REQUEST is not None:
            REQUEST.RESPONSE.write(self.mi_import_progress(REQUEST=REQUEST))
            REQUEST.RESPONSE.write('<input style="display: none;" id="inputTotal" value="%s">' % number_to_add)

        for i in xrange(len(import_maps)):
            import_map = import_maps[i]
            props = {}
            username = ''
            password = ''
            roles = default_role_is_member and ['Member'] or []
            domains = []
            for k,v in import_map.items():
                if k == 'username':
                    username = v
                elif k == 'password':
                    password = v
                elif k == 'roles':
                    roles = [s.strip() for s in re.split(r"\W", v)]
                elif k == 'domains':
                    domains = [s.strip() for s in re.split(r"\W", v)]
                else:    
                    props[k] = v            
            pm.addMember(username, password, roles, domains, props)

            # if more objects are imported than can be held in the cache
            # it slows down. So we do partial commits for every 200 objects
            if i and not i % 200:
                get_transaction().commit()

            # Output progress
            if ((i+1) % 50 == 0) and (REQUEST is not None):
                REQUEST.RESPONSE.write('<input style="display: none;" name="inputProgress" value="%s">' % (i+1))

        get_transaction().commit()


    security.declareProtected(CMFCorePermissions.View,
                              'importCSV')
    def importCSV(self, delimiter='comma', quote_char='double_quote', 
                  coding='utf-8', default_role_is_member=False,
                  REQUEST=None):
        "Convert the uploaded csv data to member types"

        pmd = getToolByName(self, 'portal_memberdata')
        reg_tool = getToolByName(self, 'portal_registration')
        
        valid_roles = pmd.validRoles()
        
        fields_map = self._getFieldsMap(REQUEST)        
        
        errors = []

        # username must always be mapped
        if not fields_map.has_key('username'):
            errors.append({'msg':"Username is a required mapping", 'row':''})

        delimiter_map = {'string':' ','text':'\n'}
        field_types = self._field_types()
        rows = self.csv_parser(delimiter, quote_char, coding)
        # get the data from the fields and put them in dict of
        # {attrname:value}
        import_maps = []
        row_idx = 1
        for row in rows:
            errorflag = 0
            import_map = {}
            for key, val in fields_map.items():
                if key.find('.') != -1:
                    key, subkey = key.split('.')
                field_type = field_types[key]
                delimiter = delimiter_map.get(field_type, ' ')
                
                vals = []
                # find the values
                row_len = len(row)
                for v in val:
                    if (v < row_len) and row[v]:
                        if field_type == 'boolean':
                            # Convert string into '1' (True) or '' (False).
                            # The value must be a string since it is an argument to join() later in this method.
                            elem = row[v]
                            b = (type(elem) is type('string')) and \
                                (elem.lower() in ('true','yes','t','y') and '1') or \
                                (elem.lower() in ('false','no','f','n') and '')
                                
                            # Any values other than those in the expression above (y,n etc) will return False. Detect this and abort.
                            if b is False: 
                                errors.append({'msg':"Row %s contains an invalid boolean value (%s)" % (row_idx, elem), 'row':','.join(row)})
                                errorflag = 1
                            
                            vals.append(b)
                            
                        elif field_type == 'integer':                         
                            try:
                                i = int(row[v])
                            except:
                                errors.append({'msg':"Row %s contains an invalid integer value (%s)" % (row_idx, row[v]), 'row':','.join(row)})
                                errorflag = 1
                                
                            vals.append(str(i))
                            
                        elif field_type == 'datetime':                         
                            from DateTime import DateTime
                            try:
                                d = DateTime(row[v])
                            except:
                                errors.append({'msg':"Row %s contains an invalid datetime value (%s)" % (row_idx, row[v]), 'row':','.join(row)})
                                errorflag = 1
                                
                            vals.append(str(d))
                            
                        elif (key == 'roles'):
                            # Check that all roles are valid                            
                            # Split on any non-alphanumerical delimiter
                            import re                            
                            roles = [s.strip() for s in re.split(r"\W", row[v])]
                            for role in roles:
                                if not role in valid_roles:    
                                    errors.append({'msg':"Row %s contains an invalid role (%s)" % (row_idx, role), 'row':','.join(row)})
                                    errorflag = 1
                                    break
                                    
                            vals.append(row[v])
                            
                        elif key == 'username':
                            # Username may not yet exist and must be valid
                            username = row[v]
                            if not reg_tool.isMemberIdAllowed(username):
                                errors.append({'msg':"Row %s contains an invalid username (%s)" % (row_idx, username), 'row':','.join(row)})
                                errorflag = 1

                            vals.append(username)

                        else:
                            vals.append(row[v])
                    else:
                        vals.append('')
                
                if not errorflag:
                    import_map[key] = delimiter.join(vals)
                        
            if not errorflag:                        
                import_maps.append(import_map)

            row_idx += 1
           
        # Errors?
        if len(errors) > 0:
            container = self.unrestrictedTraverse(
                self.REQUEST.get('current_path'))
            return container.mi_import_error(REQUEST=REQUEST, errors=errors)

        # add the members
        self._addMembers(default_role_is_member, import_maps, REQUEST)

        # delete file from session object as it can be several MB
        self._del_session_cache()
 
        if REQUEST:
            # This trick is needed since previous calls to RESPONSE.write trashes a normal redirect
            container = self.unrestrictedTraverse(
                self.REQUEST.get('current_path'))
            REQUEST.RESPONSE.write('<script>document.location.href="%s?portal_status_message=Import%%20succeeded"</script>' % container.absolute_url())

    security.declareProtected(CMFCorePermissions.View, 'base64encode')
    def base64encode(self, s):
        """ Encode a string in base64
        """
        return base64.encodestring(s)

    security.declareProtected(CMFCorePermissions.View, 'base64decode')
    def base64decode(self, s):
        """ Decode a base64 encode string
        """
        return base64.decodestring(s)

InitializeClass(MemberImportTool)
